package Sim;

/**
 * Location represents a Node's location in the World.
 *
 * @author Jason Winnebeck, Ben Willis
 */
public class Location {
	/**
	 * coordinates
	 */
	private final double x;
	private final double y;

	/**
	 * Construct a new location object, given the x,y coordinates.
	 * 
	 * @param x
	 * @param y
	 */
	public Location( double x, double y ) {
		this.x = x;
		this.y = y;
	}

	public double getX() {
		return x;
	}

	public double getY() {
		return y;
	}

	/**
	 * Returns the square of the distance between two locations, or quadrance.
	 *
	 * @param other non-null other Location
	 *
	 * @return Quadrance of a line segment between this Location and other
	 */
	public double getQuadrance( Location other ) {
		final double dx = other.x - x;
		final double dy = other.y - y;

		return dx * dx + dy * dy;
	}

	/**
	 * Get the distance between two points.
	 * @param other another location object to get the distance
	 * 				between
	 * 
	 * @return distance between input and this.
	 */
	public double getDistance( Location other ) {
		return Math.sqrt( getQuadrance( other ) );
	}

	/**
	 * Multiply the location by a scalar and return the new
	 * location object.
	 * 
	 * @param scalar number to multiple the x,y values by
	 * @return new location object with the changes
	 */
	public Location mult( double scalar ) {
		return new Location( x * scalar, y * scalar );
	}

	/**
	 * Find the angle of the triangle at this objects point using
	 * law of cosines. 
	 * @param l1
	 * @param l2
	 * @return
	 */
	public double getAngle(Location l1, Location l2){
		//c^2 = a^2 + b^2 - 2ab*cos(C)
		//	or
		//C = acos( ( c^2 - a^2 - b^2 ) / ( -2ab ) )
		double a = this.getDistance(l1);
		double b = this.getDistance(l2);
		double c = l1.getDistance(l2);
		double denom = 0;
		if(a != 0 || b != 0){
			denom = (-2*a*b);
		}
		double angle = Math.acos((Math.pow(c, 2) - Math.pow(b, 2) - Math.pow(a, 2))/denom);

		return angle;
	}
	
	/**
	 * Adds two location objects to one another.
	 * 
	 * @param other Object to add
	 * @return new location of addition of objects.
	 */
	public Location add( Location other ) {
		return new Location( x + other.x, y + other.y );
	}

	/**
	 * subtracts two location objects to one another.
	 * 
	 * @param other Object to add
	 * @return new location of addition of objects.
	 */
	public Location sub( Location other ) {
		return new Location( x - other.x, y - other.y );
	}
	
	/**
	 * Returns the size of this Location as it is seralized on the network.
	 */
	@SuppressWarnings( { "MethodMayBeStatic" } )
	public int getSize() {
		return 8;
	}

	public String toString() {
		return "{" + x + ", " + y + "}";
	}
}
