package Sim;

import Sim.Routing.RoutingAlgorithm;
import org.gillius.Util.CheckUtils;

/**
 * Node is the central object of the simulation, which tracks its location and keeps a
 * RoutingAlgorithm.
 *
 * @author Jason Winnebeck
 */
public class Node {

	//montior
	private NodeMonitor monitor = null;

	//Identity
	private final int address;
	private final Simulation sim;

	//Implementations
	private final LocationManager locationManager;
	private final RoutingAlgorithm routingAlg;

	//Attributes
	private double transmissionRange;

	//statistics variables
	private double powerUsed = 0.0;

	//Working variables
	/**
	 * Minimum time when we can send the next packet.
	 */
	private double nextSendTime = 0.0;

	public Node( int address, Simulation sim, LocationManager locationManager,
	             RoutingAlgorithm routingAlg, double transmissionRange, NodeMonitor monitor ) {
		CheckUtils.checkNull( sim, "sim" );
		CheckUtils.checkNull( locationManager, "locationManager" );
		CheckUtils.checkNull( routingAlg, "routingAlg" );

		if ( transmissionRange < 0 )
		  throw new IllegalArgumentException( "Parameter transmissionRange to Node.Node must be positive or zero, but " + transmissionRange + " was given." );


		this.address = address;
		this.sim = sim;
		this.locationManager = locationManager;
		this.routingAlg = routingAlg;
		this.transmissionRange = transmissionRange;

		routingAlg.attached( this );

		//bjw - monitoring
		if(monitor != null){
			this.monitor = monitor;
			monitor.setStartPosition(locationManager.getLocation(0.0));
		}
	}

	/**
	 * Returns the current location for this Node, based on the Simulation's current time.
	 *
	 * @return non-null Location
	 */
	public Location getLocation() {
		return locationManager.getLocation( sim.getTime() );
	}

	public int getAddress() {
		return address;
	}

	public double getTransmissionRange() {
		return transmissionRange;
	}

	public double getPowerUsed() {
		return powerUsed;
	}

	/**
	 * Occurs when a node receives a frame. Based on the contents of the frame, passes to the
	 * appropriate RoutingAlgorithm method.
	 *
	 * @param frame non-null HardwareFrame
	 */
	public void receiveFrame( HardwareFrame frame ) {
		CheckUtils.checkNull( frame, "frame" );

		if ( frame.getSource() == address ) {
			routingAlg.loopback( frame );

		} else if ( frame.isBroadcast() || frame.getDest() == address ) {
			Packet packet = frame.getUserData();
			if ( packet != null && packet.getDest() == address ) {
				//bjw - monitoring
				if(monitor != null){
					monitor.incFramesReceived();
					monitor.incPacketsTerminated();
				}
				packet.reportHop(this.getLocation().getDistance(sim.getWorld().getNode(frame.getSource()).getLocation()));
				packet.reportArrival( sim.getTime() );
				routingAlg.terminated( frame );
			} else {
				routingAlg.arrived( frame );

				//bjw - monitoring
				if(monitor != null){
					monitor.incFramesReceived();
					monitor.incPacketsArrived();
				}
				if(packet != null)
					packet.reportHop(sim.getWorld().getNode(frame.getSource()).getLocation().getDistance(this.getLocation()));
			}

		} else {
			routingAlg.overheard( frame );

			//bjw - monitoring
			if(monitor != null){
				monitor.incPacketsOverheard();
			}
		}
	}

	/**
	 * Originates a new Packet to be sent to the given destination. The addresses in the Packet must
	 * correspond properly with this Node and the destination Node. A packet origination event
	 * occurs in this method.
	 *
	 * @param destination non-null destination Node
	 * @param packet      non-null Packet to send, with proper addresses set
	 */
	public void sendNewPacket( Node destination, Packet packet ) {
		CheckUtils.checkNull( destination, "destination" );
		CheckUtils.checkNull( packet, "packet" );

		if ( address != packet.getSource() )
			throw new IllegalArgumentException( "Packet's source does not match this object" );
		if ( destination.getAddress() != packet.getDest() )
			throw new IllegalArgumentException( "Packet's dest does not match destination node" );

		packet.reportSent( sim.getTime() );
		routingAlg.newPacket( destination, packet );

		//bjw - monitoring
		if(monitor != null){
			monitor.incPacketsSent();
			monitor.addPacketsOriginated(packet);
		}
	}

	/**
	 * Originates a new Packet to be sent. The addresses in the Packet must correspond properly with
	 * this Node and a valid destination Node in the Simulation's World. A packet origination event
	 * occurs in this method.
	 *
	 * @param packet non-null Packet to send, with proper addresses set
	 */
	public void sendNewPacket( Packet packet ) {
		Node destination = sim.getWorld().getNode( packet.getDest() );
		if ( destination == null )
			throw new IllegalArgumentException( "Packet's destination address does not exist" );

		sendNewPacket( destination, packet );

		//bjw - monitoring
		if(monitor != null){
			monitor.incPacketsSent();
			monitor.addPacketsOriginated(packet);
		}
	}

	/**
	 * Queues a HardwareFrame to be sent by this node. Currently this method does not support hacking
	 * via forged frames.
	 *
	 * @param frame non-null HardwareFrame with source address equal to this node
	 */
	public void sendFrame( HardwareFrame frame ) {
		CheckUtils.checkNull( frame, "frame" );
		if ( frame.getSource() != address )
			throw new IllegalArgumentException( "HardwareFrame does not originate from this Node" );
		if ( frame.getDest() == address )
			throw new IllegalArgumentException( "Cannot send a loopback frame" );

		double packetSendTime = frame.getSizeInBits() / sim.getParams().getTransferRate();
		nextSendTime = Math.max( nextSendTime, sim.getTime() );

		//Schedule to occur at the end of the send period. Added time twice as it reads a little better.
		sim.scheduleEvent( nextSendTime + packetSendTime, new WorldSendEvent( frame ) );
		nextSendTime += packetSendTime;

		//PLEASE, take this out if this ever becomes unreasonable -- this is a sanity check to guard
		//against obnoxious flooding of packets.
//		if ( nextSendTime - sim.getTime() > 0.500 ) {
//			throw new IllegalArgumentException( "Queue full, over 0.5s of packets." );
			//I think we should just leave this out since Champ will fail without
			//it.  Due to it not working correctly.
//		}

		//bjw - monitoring
		if ( monitor != null ) {
			monitor.frameSent( frame );
			monitor.addToTotalSendTime( packetSendTime );
		}
	}

	public String toString() {
		return "Node " + address + " at " + getLocation();
	}

	/**
	 * An event that puts a frame into the world where the Node is when actually sent.
	 */
	private class WorldSendEvent implements SimulationEvent {
		private final HardwareFrame frame;

		public WorldSendEvent( HardwareFrame frame ) {
			this.frame = frame;
		}

		/**
		 * Sends the Packet to the World.
		 *
		 * @param sim  non-null Simulation
		 * @param time simulation time, exactly equal to the result of {@link Sim.Simulation#getTime()}
		 */
		public void perform( Simulation sim, double time ) {
			sim.getWorld().sendFrame( frame, getLocation(), transmissionRange );
		}
	}
}
