package Sim;

import org.gillius.Util.CheckUtils;

import java.util.Map;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.ArrayList;
/**
 * PacketWatcher watches over any Packets to which it is assigned to listen, and maintains
 * statistics on those packets.
 *
 * @author Jason Winnebeck
 */
public class PacketWatcher implements PacketListener {
	private final double timeout;

	private final Map<Packet, Double> outstanding = new IdentityHashMap<Packet, Double>();
	
	private ArrayList<Integer> hopList = new ArrayList<Integer>();
	private ArrayList<Double> distanceList = new ArrayList<Double>();
	private Map<Packet,Double> packetStart = new IdentityHashMap<Packet,Double>();
	private ArrayList<Double> timeList = new ArrayList<Double>();
	
	private int created = 0;
	private int arrived = 0;
	private int lost    = 0;
	private int hops 	= 0;
	
	/**
	 * Constructs a new PacketWatcher.
	 *
	 * @param timeout if any packet takes more than this many seconds to arrive, it is considered
	 *                lost.
	 */
	public PacketWatcher( double timeout ) {
		this.timeout = timeout;
	}

	/**
	 * Schedules a recurring event to operate in the Simulation to call {@link #updateLost(double)}
	 * every "rate" seconds, starting from the simulations current time plus "rate." The most often
	 * the event occurs, the less memory is used, but updating the statistic too often can cause a
	 * high amount of CPU overhead.
	 *
	 * @param sim  non-null Simulation
	 * @param rate event recurrence rate
	 */
	public void scheduleRecurringLossUpdate( Simulation sim, double rate ) {
		CheckUtils.checkNull( sim, "sim" );
		sim.scheduleEvent( sim.getTime() + rate,
		                   new RecurringLossUpdateEvent( rate ) );
	}

	/**
	 * Returns the number of originated packets.
	 */
	public int getCreated() {
		return created;
	}

	/**
	 * Returns the number of packets still outstanding (created packets that have yet to be lost or
	 * arrive).
	 */
	public int getOutstanding() {
		if ( outstanding.size() != created - arrived - lost )
			throw new AssertionError( "Inconsistent outstanding size" );

		return created - arrived - lost;
	}

	/**
	 * Returns number of successfully arrived packets.
	 */
	public int getArrived() {
		return arrived;
	}

	/**
	 * Returns the number of lost packets. {@link #updateLost(double)} should be called before this
	 * method to make sure that the loss statistic is up to date.
	 */
	public int getLost() {
		return lost;
	}
	
	/**
	 * Returns the list of hops for each packet.
	 * @return
	 */
	public ArrayList<Integer> getHopsList(){
		return hopList;
	}

	/**
	 * Returns the list of hops for each packet.
	 * @return
	 */
	public ArrayList<Double> getDistancesList(){
		return this.distanceList;
	}

	/**
	 * Returns the list of hops for each packet.
	 * @return
	 */
	public ArrayList<Double> getTimeList(){
		return this.timeList;
	}
	
	/**
	 * Calculates the lost packets statistic by examining the timeouts. This method can be called
	 * periodically to reduce the amount of memory used.
	 *
	 * @param time a time value to update that is greater than or equal to any time value passed to
	 *             {@link #packetOriginated(Packet, double)} or
	 *             {@link #packetArrived(Packet, double, boolean)}.
	 */
	public void updateLost( double time ) {
		//TODO: This might be a weak spot in performance, to scan the whole map every time
		for ( Iterator<Map.Entry<Packet, Double>> it = outstanding.entrySet().iterator();
		      it.hasNext(); ) {
			if ( it.next().getValue() < time ) {
				it.remove();
				++lost;
			}
		}
	}

	/**
	 * Event called when a Packet is given to a Node for sending. This event only occurs once for any
	 * given Packet.
	 *
	 * @param p    non-null Packet
	 * @param time Time of origination.
	 */
	public void packetOriginated( Packet p, double time ) {
		++created;
		if ( outstanding.put( p, time + timeout ) != null )
			throw new IllegalStateException( "Already seen " + p + " originated." );
		packetStart.put(p, time);
	}
	
	/**
	 * Event called when a Packet arrives at its destination. This method can happen multiple times if
	 * there is packet duplication. The first time a Packet arrives, repeat is false. All subsequent
	 * events, if they occur, repeat is true.
	 *
	 * @param p      non-null Packet that arrived
	 * @param time   Time of arrival, which is after time of origination for this Packet
	 * @param repeat true if another copy of the Packet arrived, false if first time.
	 */
	public void packetArrived( Packet p, double time, boolean repeat ) {
		if ( repeat ) {
			if ( outstanding.containsKey( p ) )
				throw new AssertionError( "Repeat packet is outstanding" );

		} else {
			Double val = outstanding.remove( p );
			if ( val != null ) {
				//If val is null, this packet should already have been counted as lost.

				if ( val < time ) {
					++lost; //to be consistent as if updateLost is called always
				} else {
					timeList.add( time - packetStart.get(p) );
					++arrived;
				}
			}
		}
		//TODO: consider counting packet duplication?
	}

	public void reportTotalHops(Packet p, int hops, double d){
		hopList.add(hops);
		distanceList.add(d);
	}
	
	private class RecurringLossUpdateEvent implements SimulationEvent {
		private double rate;

		public RecurringLossUpdateEvent( double rate ) {
			this.rate = rate;
		}

		/**
		 * Performs the event that is occuring within a {@link Sim.Simulation}. If an event wishes to
		 * re-execute later, it may schedule itself again using the provided Simulation reference.
		 *
		 * @param sim  non-null Simulation
		 * @param time simulation time, exactly equal to the result of {@link Sim.Simulation#getTime()}
		 */
		public void perform( Simulation sim, double time ) {
			updateLost( time );
			sim.scheduleEvent( time + rate, this );
		}
	}

}
