package Sim.Routing;

import Sim.Node;
import Sim.HardwareFrame;
import Sim.Packet;
import Sim.Simulation;
import Sim.SimulationEvent;
import org.gillius.Util.CheckUtils;

/**
 * AbstractRoutingAlgorithm is a RoutingAlgorithm implementation whose methods take no action, so
 * that only the methods of interest can be overridden.
 *
 * @author Jason Winnebeck
 */
public class AbstractRoutingAlgorithm implements RoutingAlgorithm {
	private   Simulation sim;
	protected Node       myNode;

	/**
	 * Sets the simulation for this RoutingAlgorithm. This method must be called immediately after
	 * construction and before any other methods.
	 *
	 * @param sim non-null Simulation
	 */
	public void setSimulation( Simulation sim ) {
		CheckUtils.checkNull( sim, "sim" );
		this.sim = sim;
	}

	/**
	 * Event called when a Node is given a RemoteAlgorithm to route packets for it.
	 *
	 * @param us non-null Node
	 */
	public void attached( Node us ) {
		myNode = us;
	}

	/**
	 * Event when a HardwareFrame arrives destined for this Node that does not have a user Packet, or
	 * has a Packet not for this Node. The data in the frame may need to be routed.
	 *
	 * @param in non-null HardwareFrame received
	 */
	public void arrived( HardwareFrame in ) {
	}

	/**
	 * Event when a HardwareFrame arrives destined for this Node that has a user Packet that is
	 * destined for this Node. The data has arrived at its final destination, so it does not need to be
	 * routed.
	 *
	 * @param in non-null HardwareFrame received
	 */
	public void terminated( HardwareFrame in ) {
	}

	/**
	 * Event when a HardwareFrame is overheard by a Node (the HardwareFrame's destination is not the
	 * Node).
	 *
	 * @param in non-null HardwareFrame received
	 */
	public void overheard( HardwareFrame in ) {
	}

	/**
	 * Event when a Node hears its own HardwareFrame (which is guaranteed to happen after a send
	 * completes).
	 *
	 * @param in non-null HardwareFrame received whose source equals the "us" address
	 */
	public void loopback( HardwareFrame in ) {
	}

	/**
	 * Event when a Node wants to send a new Packet. The RoutingAlgorithm should bundle the Packet in a
	 * HardwareFrame and send it through the Node.
	 *
	 * @param dest   non-null Node as the destination (packet.getDest() == dest.getAddress())
	 * @param packet user packet data to send, with a destination address not equal to the address for
	 *               the sending Node.
	 */
	public void newPacket( Node dest, Packet packet ) {
	}

	/**
	 * Returns the current time of the Simulation.
	 *
	 * @return non-null Simulation's time
	 */
	protected double getTime() {
		return sim.getTime();
	}

	/**
	 * Schedules a new event to take place in the future. Once an event has been scheduled, it
	 * cannot be cancelled or rescheduled.
	 *
	 * @param t     event time greater than or equal to the current simulation time
	 * @param event non-null {@link SimulationEvent}
	 */
	public void scheduleEvent( double t, SimulationEvent event ) {
		sim.scheduleEvent( t, event );
	}
}
