package Sim.Routing.DSR;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

/**
 * Route provides a convenient interface for dealing with routing chains with copy-on-write
 * semantics. Route has a copy-on-write semantic so that any modification to a Route returns a new
 * Route while leaving the original object unmodified. Therefore, a Route instance is immutable.
 * This is the same technique used by java.lang.String.
 *
 * @author Jason Winnebeck
 */
public class Route {
	private final List<Integer> route;

	private Route( List<Integer> route ) {
		this.route = route;
	}

	public Route( Integer... path ) {
		//copy to prevent modification
		route = new ArrayList<Integer>( Arrays.asList( path ) );
	}

	public int numHops() {
		return route.size();
	}

	public boolean containsAddress( int address ) {
		return route.contains( address );
	}

	/**
	 * Returns the Route from the node with the given address onwards to the end, or null if the given
	 * node is not in Route. The returned Route has "address" as its first hop, and the last hop is
	 * the same as the last hop from this Route.
	 */
	public Route getRouteFrom( int address ) {
		int idx = route.indexOf( address );
		if ( idx >= 0 )
			return new Route( route.subList( idx, route.size() ) );
		else
			return null;
	}

	/**
	 * Returns true if there is a link from 'start' to 'end' in this Route.
	 */
	public boolean containsLink( int start, int end ) {
		for ( int i = 0; i < route.size() - 1; i++ ) {
			if ( route.get( i ) == start && route.get( i + 1 ) == end )
				return true;
		}
		return false;
	}

	/**
	 * Adds a hop to the end of this Route.
	 */
	public Route addHop( int address ) {
		List<Integer> newRoute = new ArrayList<Integer>( route.size() + 1 );
		newRoute.addAll( route );
		newRoute.add( address );
		return new Route( newRoute );
	}

	public int getFirstHop() {
		if ( route.isEmpty() )
			throw new IllegalStateException( "There is no first hop" );

		return route.get( 0 );
	}

	/**
	 * Returns the hop following the first hop.
	 */
	public int getSecondHop() {
		if ( route.size() < 2 )
			throw new IllegalStateException( "There is no second hop" );

		return route.get( 1 );
	}

	public int getLastHop() {
		if ( route.isEmpty() )
			throw new IllegalStateException( "There is no last hop" );

		return route.get( route.size() - 1 );
	}

	public Route removeFirstHop() {
		if ( route.isEmpty() )
			throw new IllegalStateException( "There is no first hop" );

		return new Route( route.subList( 1, route.size() ) );
	}

	public Route removeLastHop() {
		if ( route.isEmpty() )
			throw new IllegalStateException( "There is no last hop" );

		return new Route( route.subList( 0, route.size() - 1 ) );
	}

	public Route reverse() {
		List<Integer> newRoute = new ArrayList<Integer>( route );
		Collections.reverse( newRoute );
		return new Route( newRoute );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		Route other = (Route) o;
		return route.equals( other.route );
	}

	public int hashCode() {
		return route.hashCode();
	}

	public String toString() {
		return "Route " + route;
	}

	/**
	 * Returns the size of this Route if it were to be serialized as bytes.
	 */
	public int getSize() {
		return 1 + numHops() * 4;
	}
}
