package Sim.Routing;

import Sim.HardwareFrame;
import Sim.Node;
import Sim.Packet;
import Sim.ProtocolFrame;

import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Iterator;

/**
 * FloodingRoutingAlgorithm implements a simple routing algorithm where every user data packet is
 * flooded to all nodes. A cache of seen packets prevents reflooding of the same packet (which
 * prevents loops in the routing).
 *
 * @author Jason Winnebeck
 */
public class FloodingRoutingAlgorithm extends AbstractRoutingAlgorithm {
	/**
	 * Packet ID to non-null expiry time.
	 */
	private Map<PacketId, Double> seen = new IdentityHashMap<PacketId, Double>();

	private static final Double EXPIRE_TIME = 3.0;

	/**
	 * Event when a Node wants to send a new Packet. The RoutingAlgorithm should bundle the Packet in a
	 * HardwareFrame and send it through the Node.
	 *
	 * @param dest   non-null Node as the destination (packet.getDest() == dest.getAddress())
	 * @param packet user packet data to send, with a destination address not equal to the address for
	 *               the sending Node.
	 */
	@Override
	public void newPacket( Node dest, Packet packet ) {
		PacketId pid = new PacketId();
		HardwareFrame frame = new HardwareFrame(
				myNode.getAddress(), HardwareFrame.BROADCAST, new Header( 10, pid ), packet );
		seen.put( pid, getTime() + EXPIRE_TIME );

		myNode.sendFrame( frame );
	}

	/**
	 * Event when a HardwareFrame arrives destined for this Node that does not have a user Packet, or
	 * has a Packet not for this Node. The data in the frame may need to be routed.
	 *
	 * @param in non-null HardwareFrame received
	 */
	@Override
	public void arrived( HardwareFrame in ) {
		doExpires( getTime() );
		Header header = (Header) in.getProtocolData();

		if ( !seen.containsKey( header.id ) ) {
			seen.put( header.id, getTime() + EXPIRE_TIME );
			Header newHeader = header.decrement();
			if ( newHeader != null ) {
				HardwareFrame frame = new HardwareFrame(
						myNode.getAddress(), HardwareFrame.BROADCAST, newHeader, in.getUserData() );
				myNode.sendFrame( frame );
			}
		}
	}

	/**
	 * Run through the cache map and removes expired entries (entries older than the given time).
	 */
	private void doExpires( double time ) {
		//TODO: This might be a weak spot in performance, to scan the whole map every time
		//Maybe we could use a parallel priority queue (or sorted circular buffer)?
		for ( Iterator<Map.Entry<PacketId, Double>> it = seen.entrySet().iterator(); it.hasNext(); ) {
			if ( it.next().getValue() < time ) {
				it.remove();
			}
		}
	}

	private static class Header implements ProtocolFrame {
		private final int      hopsRemaining;
		public  final PacketId id;

		public Header( int hopsRemaining, PacketId id ) {
			this.hopsRemaining = hopsRemaining;
			this.id = id;
		}

		public Header decrement() {
			if ( hopsRemaining > 0 )
				return new Header( hopsRemaining - 1, id );
			else
				return null;
		}

		/**
		 * Returns the "serialized" size of this portion of the packet.
		 */
		public int getSize() {
			return 1 + 4; //1 for hops, 4 for unique ID (we don't need to count origSource.
		}
	}

	private static class PacketId {
	}
}
