package Sim.Routing.GPSR;

import Sim.HardwareFrame;
import Sim.Node;
import Sim.Packet;
import Sim.Location;
import Sim.Routing.GPSR.Header.*;
import Sim.Simulation;

import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import Sim.Line;
/**
 * Greedy Perimeter Stateless Routing (GPSR) Algorithm, based on nodes know their own locations. 
 * Uses a broadcast technique to keep 1 hop nodes up to date with locations of neighbors.  Packet
 * sending is done based on greedy, or perimeter routing.  Greedy uses the node nearest to the 
 * destination location to route the packet.  Because there isn't always a node closer than the
 * originator, perimeter routing attempts to push the packet around voids in a right hand rule
 * method.  
 *
 * @author Benjamin Willis
 */
public class Algorithm extends Sim.Routing.AbstractRoutingAlgorithm {
	/**
	 * Packet ID to non-null expiry time.
	 */
	private Map<PacketId, Double> seen = new IdentityHashMap<PacketId, Double>();
	/**
	 * Map of neighbors, contains address and location of the neighbor.
	 */
	private Map<Integer, ExpiringLocation> neighbors = new HashMap<Integer, ExpiringLocation>();
	/**
	 * Time for packets seen to expire, non algorithm related.
	 */
	private static final Double EXPIRE_TIME = 1000.0;
	/**
	 * modes for which a packet can be in
	 */
	private static final int MODE_GREEDY = 0;
	private static final int MODE_PERIMETER = 1;
	
	/**
	 * Configures initial events to start when the simulation starts.
	 * @param sim the simulation to run with this algorithm
	 */
	public void setSimulation( Simulation sim ) {
		super.setSimulation( sim );
		sim.scheduleEvent( getTime() + myNode.getAddress()*.1, new BroadcastEvent(myNode) );
	}
	
	/**
	 * Event when a Node wants to send a new Packet. The RoutingAlgorithm should bundle the Packet in a
	 * HardwareFrame and send it through the Node.
	 *
	 * @param src    non-null Node originating the Packet (packet.getSource() == src.getAddress())
	 * @param dest   non-null Node as the destination (packet.getDest() == dest.getAddress())
	 * @param packet user packet data to send, with a destination address not equal to the address for
	 *               the sending Node.
	 */
	@Override
	public void newPacket( Node dest, Packet packet ){
		//System.out.println("newPacket from " + myNode.getAddress() + " to " + dest.getAddress());
		PacketId pid = new PacketId();
		//make a new hardwareframe to wrap the packet so i can send it out
		//make a new header
		if(neighbors.size() == 0){
		//	System.out.println("no neighbors known");
		}else{
			int destAddress = getGreedyNode(
					myNode.getAddress(),
					myNode.getLocation(),
					dest.getAddress(),
					dest.getLocation());
			
			HardwareFrame frame;
			if(destAddress == myNode.getAddress()){
				//start perimeter routing
				destAddress = getPerimeterNode(myNode.getAddress(),
						myNode.getLocation(),
						dest.getLocation());
				frame = new HardwareFrame(
						myNode.getAddress(), 
						destAddress, 
						new Standard(pid,
								MODE_PERIMETER,
								myNode.getAddress(),
								myNode.getLocation(),
								dest.getLocation(),
								myNode.getLocation()),
						packet );
			}else{
				//start greedy routing
				frame = new HardwareFrame(
						myNode.getAddress(), 
						destAddress, 
						new Standard(pid,
								MODE_GREEDY,
								myNode.getAddress(),
								myNode.getLocation(),
								dest.getLocation()),
						packet );
			}
			//System.out.println("{"+myNode.getAddress()+"} forwarding to {"+destAddress+"}");
			if(destAddress != myNode.getAddress()){
				seen.put( pid, getTime() + EXPIRE_TIME );
				myNode.sendFrame( frame );
			}else{
				//drop
			}
		}
	}

	/**
	 * Event when a HardwareFrame arrives destined for this Node that does not have a user Packet, or
	 * has a Packet not for this Node. The data in the frame may need to be routed.
	 *
	 * @param us non-null Node receiving the HardwareFrame
	 * @param in non-null HardwareFrame received
	 */
	@Override
	public void arrived( HardwareFrame in ) {
		
		doExpires(getTime());
		HardwareFrame frame = null;
		if(in.isBroadcast()){
			Broadcast header = (Broadcast) in.getProtocolData();
			if ( !seen.containsKey( header.id )) {
				seen.put( header.id, getTime() + EXPIRE_TIME );
				addOrUpdateNeighbor(header,getTime());
			}
		}else{
			Standard header = (Standard) in.getProtocolData();
			Packet p = in.getUserData();
			if ( !seen.containsKey( header.id )) {
				
				seen.put( header.id, getTime() + EXPIRE_TIME );
				addOrUpdateNeighbor(header,getTime());
				int destAddress = getGreedyNode(
						myNode.getAddress(),
						myNode.getLocation(),
						p.getDest(),
						header.getFinalLocation());
				
				if(destAddress == myNode.getAddress()){
//					System.out.println("Perimeter forwarding..");
					//start new perimeter mode
					destAddress = getPerimeterNode(myNode.getAddress(),
							myNode.getLocation(),
							header.getFromLocation());
					
					frame = new HardwareFrame(
							myNode.getAddress(), 
							destAddress, 
							new Standard(header.id,
									MODE_PERIMETER,
									myNode.getAddress(),
									myNode.getLocation(),
									header.getFinalLocation(),
									myNode.getLocation()
									),
							in.getUserData() );
				}else if(header.getMode() == MODE_PERIMETER && 
						(myNode.getLocation().getDistance(header.getFinalLocation()) >
						header.getPerimeterStarted().getDistance(header.getFinalLocation()))){
	//				System.out.println("continuing Perimeter forwarding");

					destAddress = getPerimeterNode(myNode.getAddress(),
							myNode.getLocation(),
							header.getFromLocation());
					//continue perimeter mode
					frame = new HardwareFrame(
							myNode.getAddress(), 
							destAddress, 
							new Standard(header.id,
									MODE_PERIMETER,
									myNode.getAddress(),
									myNode.getLocation(),
									header.getFinalLocation(),
									header.getPerimeterStarted()),
							in.getUserData() );
				}else{
					if(header.getMode() == MODE_PERIMETER){
		//				System.out.println("continuing Greedy forwarding");
					}
					//greedy mode
					frame = new HardwareFrame(
							myNode.getAddress(), 
							destAddress, 
							new Standard(header.id,
									MODE_GREEDY,
									myNode.getAddress(),
									myNode.getLocation(),
									header.getFinalLocation()),
							in.getUserData() );
					
				}
				if(destAddress != myNode.getAddress())
					myNode.sendFrame( frame );
			//	System.out.println("{"+myNode.getAddress()+"} forwarding to {"+destAddress+"}");
			}
		}
		
	}
	

	/**
	 * Run through the cache map and removes expired entries.
	 * @param time the current time
	 */
	private void doExpires(double time) {
		//TODO: implement. Technically this is not required, but we could run out of memory.
		for ( Iterator<Map.Entry<PacketId, Double>> it = seen.entrySet().iterator(); it.hasNext(); ) {
			if ( it.next().getValue() < time ) {
				it.remove();
			}
		}
	}

	/**
	 * Action when a packet is overheard.  Only thing we need to do is update our
	 * neighbor table, these are packets not destined for us.
	 * 
	 * @param in the hardwareframe that is overheard
	 */
	public void overheard( HardwareFrame in ){
		Standard header = (Standard) in.getProtocolData();
		addOrUpdateNeighbor(header,getTime());
		//get location information and update tables
	}
	
	/**
	 * This method takes a header of any packet seen and removes the source address
	 * and location to add or update its neighbor table.
	 * 
	 * @param header input header of the frame heard.
	 */
	private void addOrUpdateNeighbor(Base header, double time){
		ExpiringLocation nearby = neighbors.get(header.getFromAddress());
		if(nearby == null){
			neighbors.put(header.getFromAddress(), new ExpiringLocation(time,header.getFromLocation()));
		}else{
			nearby.setLoc(header.getFromLocation(), time);
		}		
	}
	/**
	 * Action when a packet has finished it's journey.
	 * 
	 * @param in hardwareframe that is at it's destination.
	 */
	public void terminated( HardwareFrame in ) {
	//	System.out.println("done at "+in.getDest());	
	}

	/**
	 * Determines the node that is closest to the src and dest.
	 * 
	 * @param address address of current node
	 * @param src the initial location of the frame
	 * @param dst the destination of the frame
	 * @return address of the next node to go to
	 */
	private int getGreedyNode(int address, Location src, int destAddress, Location dst){
		double minDistance = src.getDistance(dst);
		int bestAddress = address;
		double thisDistance;
		for ( Iterator<Map.Entry<Integer, ExpiringLocation>> it = neighbors.entrySet().iterator(); it.hasNext(); ) {
			Map.Entry entry = it.next();
			if((Integer)entry.getKey() == destAddress){
				bestAddress = destAddress;
				break;
			}else{
				Location loc = ((ExpiringLocation) entry.getValue()).getLoc();
				thisDistance = dst.getDistance(loc);
				if( minDistance > thisDistance ){
					minDistance = thisDistance;
					bestAddress = (Integer) entry.getKey();
				}
			}
		}
		return bestAddress;
	}
	
	/**
	 * Determine the next perimeter node of this node given it's neighbors
	 * and previous location.
	 * @param address address of current node
	 * @param src the initial location of the frame
	 * @param dst the destination of the frame
	 * @return address of the next node to go to
	 */
	private int getPerimeterNode(int address, Location src, Location dst){
		//need to find the next line segment
		double minAngle = 360;
		double angle, x, y;
		int bestAddress = address;
		Line centerToPoint;
		
		for ( Iterator<Map.Entry<Integer, ExpiringLocation>> it = neighbors.entrySet().iterator(); it.hasNext(); ) {
			Map.Entry entry = it.next();
			ExpiringLocation n = (ExpiringLocation)entry.getValue();	
			if(!n.getLoc().equals(dst)){
				angle = src.getAngle(dst,n.getLoc());
				//Circle Parametric Equations
				//x(t) = r cos(t) + j
				//y(t) = r sin(t) + k
				x = src.getDistance(dst) * Math.cos(angle) + src.getX();
				y = src.getDistance(dst) * Math.sin(angle) + src.getY();
				centerToPoint = new Line(src,new Location(x,y));
				if(!centerToPoint.isOnLine(n.getLoc())){
					angle += Math.PI;
				}
				if(minAngle > angle){
					bestAddress = (Integer) entry.getKey();
					minAngle = angle;	
				}
			}
		}
		return bestAddress;
	}
}
