package Sim.Routing.GPSR.Header;

import Sim.Location;
import Sim.Routing.GPSR.PacketId;
/**
 * This is the basic header that is needed by GPSR.  It contains two modes which will use more or
 * less fields.  Mainly, for greedy routing, only the mode (obviously), toAddress, fromAddress and
 * fromLocation (for updating purposes), and the finalLocationa are used.
 * @author Benjamin Willis
 *
 */
public class Standard implements Base {
	public  final PacketId id;
	
	/**
	 * for neighbor updates
	 */
	private Location fromLocation;
	private int fromAddress;
	
	private Location finalLocation;
	private Location perimeterStarted = null;
	private Location faceLocation = null;
	private double edge = 0.0;
	private int mode;
	/**
	 * The constructor to create a new frame for sending a packet to a destination.
	 * 
	 * @param id a unique id for this frame.
	 * @param mode a field indicating if the packet is traveling in greedy or perimeter mode.
	 * @param toAddress the next node this header is intended for.
	 * @param fromAddress the address of node this was sent from, purely for updating neighbors.
	 * @param fromLocation the location of the sender node, purely for updating neighbors.
	 * @param finalLocation the end location that this is to be sent to.
	 */
	public Standard( PacketId id, 
			int mode, 
			int fromAddress, 
			Location fromLocation, 
			Location finalLocation) {
		this.id = id;
		this.mode = mode;
		this.fromLocation = fromLocation;
		this.fromAddress = fromAddress;
		this.finalLocation = finalLocation;
	}

	public Standard( PacketId id, 
			int mode, 
			int fromAddress, 
			Location fromLocation, 
			Location finalLocation,
			Location perimeterStarted) {
		this.id = id;
		this.mode = mode;
		this.fromLocation = fromLocation;
		this.fromAddress = fromAddress;
		this.finalLocation = finalLocation;
		this.perimeterStarted = perimeterStarted;
	}
	
	/**
	 * Returns the "serialized" size of this portion of the packet.
	 */
	public int getSize() {
		return 5*8; //1 for hops, 4 for unique ID (we don't need to count origSource.
	}
	
	/**
	 * The address of the node that sent this header.
	 * @return int of address of sender.
	 */
	public int getFromAddress() {
		return fromAddress;
	}
	
	/**
	 * The location of the node that sent this header.
	 * @return location object of coordinates of originator node.
	 */
	public Location getFromLocation() {
		return fromLocation;
	}

	/**
	 * The end destination of the node that is to receive the packet.
	 * @return location object of the final node.
	 */
	public Location getFinalLocation() {
		return finalLocation;
	}

	/**
	 * Return the mode that this packet is currently routing in.
	 * @return integer of the mode.
	 */
	public int getMode() {
		return mode;
	}
	public Location getPerimeterStarted(){
		return perimeterStarted;
	}

	public Location getFaceLocation() {
		return faceLocation;
	}

	public void setFaceLocation(Location faceLocation) {
		this.faceLocation = faceLocation;
	}

	public double getEdge() {
		return edge;
	}

	public void setEdge(double edge) {
		this.edge = edge;
	}
}