/**
 *
 */
package Sim;

import Sim.Routing.AbstractRoutingAlgorithm;
import Sim.Routing.RoutingAlgorithm;
import edu.rit.numeric.prob.DoublePrng;
import edu.rit.numeric.prob.ExponentialPrng;
import org.gillius.Util.CheckUtils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;
/**
 * This class has a static method to create a simulation from a give
 * input file.  Expects a file in the format:
 * Line 1: routing_algorithm number_of_nodes transmission_range idle_power
 *   transmit_power transfer_rate noise_rate duration seed node_area
 *   location_is_random_else_uniform
 * Line 2: movement_is_random_else_specify
 * Line 3: communication_is_random_else_specify
 *
 * @author Benjamin Willis
 */
public class SimulationBuilder {

	private final static int firstLineParams = 13;

	// First line of input file
	private final static int algorithm = 0;
	private final static int numberOfNodes = 1;
	private final static int range = 2;
	private final static int idlePower = 3;
	private final static int transmitPower = 4;
	private final static int transferRate = 5;
	private final static int noiseRate = 6;
	private final static int duration = 7;
	private final static int seed = 8;
	private final static int nodeArea = 9;
	private final static int nodeLocations = 10;
	private final static int packetSize = 11;
	private final static int movementSpeed = 12;
	private final static int movementPos = 0;
	private final static int movementNode = 1;
	private final static int movementStartX = 2;
	private final static int movementStartY = 3;
	private final static int movementEndX = 4;
	private final static int movementEndY = 5;
	private final static int communication = 0;
	private final static int commSource = 1;
	private final static int commDest = 2;


	private final static int noMovement = 0;
	private final static int randomMovement = 1;
	private final static int fixedMovement = 2;

	public SimulationBuilder(){
	}
	/**
	 * Static method which sets up the entire simulation, returns a simulation runner object.
	 * 
	 * @param input single file to exe
	 * @param outputDir
	 * @return
	 * @throws IOException
	 * @throws SimulationException
	 */
	public static SimulationRunner buildSimulation(File input, File outputDir)
		throws IOException, SimulationException{

        CheckUtils.checkNull( input, "input" );
        CheckUtils.checkNull( outputDir, "outputDir" );

        if ( !input.isFile() || !outputDir.isDirectory() )
            throw new SimulationException( "Invalid input file or output directory path" );

        BufferedReader in = new BufferedReader( new FileReader( input ) );

        String nextLine = in.readLine();
        if ( nextLine == null )
            throw new SimulationException( "Must be at least 1 line in input file" );

        String[] parameters = nextLine.split( " " );

        if ( parameters.length != firstLineParams )
            throw new SimulationException( "First line must have " + firstLineParams + " parameters" );

        String rAlgorithm = parameters[algorithm];


        // Setup the simulation parameters from the file
        SimulationParameters params = new SimulationParameters(
            Double.parseDouble( parameters[idlePower] ),
            Double.parseDouble( parameters[transmitPower] ),
            Double.parseDouble( parameters[transferRate] ),
            Double.parseDouble( parameters[noiseRate] ),
            Double.parseDouble( parameters[duration] ),
            Integer.parseInt( parameters[seed] )
        );

        Simulation sim = new Simulation( params );

        //Get the transmissions and number of nodes
		Double transmissionRange = Double.parseDouble( parameters[range] );
		Integer totalNodes = Integer.parseInt( parameters[numberOfNodes] );
		Integer areaSize = Integer.parseInt( parameters[nodeArea] );
		Integer locations = Integer.parseInt( parameters[nodeLocations] );
		Integer packet = Integer.parseInt( parameters[packetSize] );
		Double nodeVelocity = Double.parseDouble( parameters[movementSpeed] );

//		 line 2 movement, 1 for random, 0 for other
		nextLine = in.readLine();
        if ( nextLine == null )
            throw new SimulationException( "Must be a 2nd line in input file" );
		parameters = nextLine.split( " " );

		Integer movement = Integer.parseInt( parameters[movementPos] );
		Integer nodeAddress=0,nodeStartX=0,nodeStartY=0,nodeEndX=0,nodeEndY=0;
		if(movement.equals(fixedMovement))
			{
			nodeAddress = Integer.parseInt( parameters[movementNode] );
			nodeStartX = Integer.parseInt( parameters[movementStartX] );
			nodeStartY = Integer.parseInt( parameters[movementStartY] );
			nodeEndX = Integer.parseInt( parameters[movementEndX] );
			nodeEndY = Integer.parseInt( parameters[movementEndY] );
			}
		LocationGenerator locgen;
		if(locations.equals(1)){
			locgen = new RandomLocationGenerator(params.getRandomSeed(),areaSize);
		}else{
			locgen = new UniformLocationGenerator(totalNodes,transmissionRange,areaSize);
		}

		Location location;
		LocationManager lm = null;
		Node node;
		RoutingAlgorithm ra = new AbstractRoutingAlgorithm();
		ArrayList<NodeMonitor> nodeMonitors = new ArrayList<NodeMonitor>();
		NodeMonitor mon;
		for(int i=0;i<totalNodes;i++){
			location = locgen.getNext();
			if(movement.equals(fixedMovement))
				{
				if(nodeAddress == i)
					{
					lm = new SetLocationManager(nodeVelocity,
							new Location(nodeStartX,nodeStartY),
							new Location(nodeEndX,nodeEndY));
					}
				else
					{
					lm = new FixedLocationManager( location );
					}
				}
			else if(movement.equals(randomMovement))
				{
				lm = new RandomLocationManager(nodeVelocity,areaSize,location,params.getRandomSeed());
				}
			else if(movement.equals(noMovement))
				{
				lm = new FixedLocationManager( location );
				}
	        //read in the routing algorithm
	        try {
	        	ra = (RoutingAlgorithm) Class.forName( rAlgorithm ).newInstance();
	        } catch ( Exception e ) {
	        	System.out.println(e.toString());
	            throw new SimulationException( "Could not create the routing algorithm", e );
	        }
	        mon = new NodeMonitor();
	        nodeMonitors.add(mon);
        	node = new Node( i, sim, lm, ra, transmissionRange, mon);
        	sim.getWorld().addNode( node );
        	ra.setSimulation(sim);
			//System.out.println("#"+i+") at (" + node.getLocation().getX() + "," + node.getLocation().getY() + ")");
		}

//		 line 3 communication, 1 for random, 0 for other
		nextLine = in.readLine();
        if ( nextLine == null )
            throw new SimulationException( "Must be a 3nd line in input file" );
		parameters = nextLine.split( " " );
		Integer randomCommunication = Integer.parseInt( parameters[communication] );

		TrafficGenerator traffgen;
		DoublePrng num = new ExponentialPrng( new Random(seed), 10 );
		PacketListener pl = new PacketWatcher(10);
		if(randomCommunication.equals(1))
			{
			traffgen = new TrafficGenerator(
					sim,
					num,
					packet,
					2.0,
					pl);
			}
		else
			{
			Node source = null;
			Node dest = null;
			source = sim.getWorld().getNode(Integer.parseInt(parameters[commSource]));
			dest = sim.getWorld().getNode(Integer.parseInt(parameters[commDest]));
			traffgen = new TrafficGenerator(
					sim,
					source,
					dest,
					packet,
					2.0,
					num,
					pl);
			}
		nextLine = in.readLine();
        if ( nextLine != null )
        	sim.setName(nextLine);

		return new SimulationRunner( sim , nodeMonitors ,(PacketWatcher) pl, outputDir );
	}
}
