package Sim;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;


/**
 * This class executes a simulation by executing the events.  Once complete, this 
 * will compile the results and output them to an xml file.
 * 
 * @author Benjamin Willis
 *
 */
public class SimulationRunner {
	private Simulation sim;
	private ArrayList<NodeMonitor> monitor;
	private int totalPacketsSent = 0, totalBytes = 0, totalPacketsTerminated = 0, totalFramesRec = 0, totalFramesSent = 0, totalPacketsProcessed = 0, totalNodes = 0, minHopLength = -1, totalHops = 0;
	private double totalSendTime = 0, percentLoss,avgSendTime,avgBytes,avgPacketsProcessed, frameLoss, powerUsage, averageHopLength, minDist=-1, avgDist=0, ttlDist=0, minTime=-1, ttlTime=0, avgTime=0;
	private int totalProtocolSent = 0, totalUserSent = 0;
	private Document xd;
	private File xmlFile;
	private PacketWatcher pl;
	
	/**
	 * Constructor to create a new simulation, after this is created, simply call <tt>runSimulation</tt>
	 * @param sim simulation to run
	 * @param monitor object that watches the simulation
	 * @param pl object that watches the packets during the simulation
	 * @param outputDirectory directory to store the xml files
	 */
	public SimulationRunner(Simulation sim,
			ArrayList<NodeMonitor> monitor,
			PacketWatcher pl,
			File outputDirectory){
		this.sim = sim;
		this.monitor = monitor;
		this.pl = pl;

		try{
			xmlFile = new File( outputDirectory, sim.getName() + ".xml" );
			DocumentBuilderFactory xmlBuilderFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder xmlBuilder = xmlBuilderFactory.newDocumentBuilder();
			xd = xmlBuilder.newDocument();
		}catch (Exception e){
			System.out.println("Couldn't construct the document object for output.");
		}
	}

	public void runSimulation() throws SimulationException{
		while ( sim.getTime() < sim.getParams().getSimDuration() ) {
			//System.out.println(sim.getTime()/sim.getParams().getSimDuration() * 100 + "%");
            if ( !sim.next() ) {
                throw new SimulationException( "Not enough events to run the simulation " +
                                               "to completion" );
            }
        }
		System.out.println("Simulation is Complete.");
		collectData();
		try{

			Element root = xd.createElement("simulation");
			root.appendChild(parameters());
			root.appendChild(results());
			xd.appendChild(root);

			writeFile();

		}catch (Exception e){
			System.out.println("Couldn't write output.");
			System.out.println(e.toString());
		}
	}
	
	/**
	 * This method collects data from the monitor and packet watcher and create store in variables the 
	 * data about the simulation.
	 */
	private void collectData(){
		Iterator iter = monitor.iterator();
		NodeMonitor currentNode;
		totalNodes = monitor.size();
		while(iter.hasNext()){
			currentNode = (NodeMonitor) iter.next();
			//totalPacketsSent += currentNode.getPacketsSent();
			totalBytes += currentNode.getTotalBytes();
			totalSendTime += currentNode.getTotalSendTime();
			//totalPacketsTerminated += currentNode.getPacketsTerminated();
			totalFramesSent += currentNode.getFramesSent();
			totalProtocolSent += currentNode.getTotalProtocolBytes();
			totalUserSent += currentNode.getTotalUserBytes();
			totalFramesRec += currentNode.getFramesReceived();
			totalPacketsProcessed += currentNode.getPacketsArrived();
		}
		totalPacketsSent = pl.getCreated();
		totalPacketsTerminated = pl.getArrived();
		percentLoss = ( 1 - (double) totalPacketsTerminated / (double) totalPacketsSent ) * 100;
		avgSendTime = totalSendTime / totalNodes;
		avgBytes = totalBytes / totalNodes;
		avgPacketsProcessed = totalPacketsProcessed / totalNodes;
		frameLoss = ( 1 - totalFramesRec / totalFramesSent ) * 100;
		powerUsage = totalSendTime * sim.getParams().getTransmitPowerUse() + (sim.getParams().getSimDuration() - sim.getParams().getTransmitPowerUse() * sim.getParams().getIdlePowerUse());
		Iterator itera = pl.getHopsList().iterator();
		Integer hops;
		while(itera.hasNext()){
			hops = (Integer) itera.next();
			if(hops < minHopLength || minHopLength == -1){
				minHopLength = hops;
			}
			totalHops += hops;
		}
		averageHopLength = totalHops / (double) pl.getArrived();

		Iterator iterat = pl.getDistancesList().iterator();
		Double dist;
		while(iterat.hasNext()){
			dist = (Double) iterat.next();
			if(dist < minDist || minDist == -1){
				minDist = dist;
			}
			ttlDist += dist;
		}
		avgDist = ttlDist / pl.getArrived();

		Iterator iterato = pl.getTimeList().iterator();
		Double time;
		while(iterato.hasNext()){
			time = (Double) iterato.next();
			if(time < minTime || minTime == -1){
				minTime = time;
			}
			ttlTime += time;
		}
		avgTime = ttlTime / pl.getArrived();

	}

	/**
	 * Writes the dom document to the output file.
	 *
	 */
	private void writeFile(){
		try{
			TransformerFactory tf = TransformerFactory.newInstance();
			Transformer trans = tf.newTransformer();
			DOMSource DOM = new DOMSource(xd);

			StreamResult SR = new StreamResult(xmlFile);
			trans.transform(DOM, SR);
		}catch (Exception e){
			System.out.println("Error writing to output file.");
		}

	}
	
	/**
	 * Parses input parameters for setup into a dom element.
	 *
	 * @return the constructed element.
	 */
	private Element parameters(){
		Element setup = xd.createElement("parameters");
		Element name = xd.createElement("name");
		Element noiseRate = xd.createElement("noiseRate");
		Element randomSeed = xd.createElement("randomSeed");
		Element duration = xd.createElement("duration");
		Element transferRate = xd.createElement("transferRate");
		Element transmitPowerUse = xd.createElement("transmitPowerUse");
		Element idlePowerUse = xd.createElement("idlePowerUse");
		Element ttlNodes = xd.createElement("totalNodes");

		name.appendChild(xd.createTextNode(String.valueOf(sim.getName())));
		noiseRate.appendChild(xd.createTextNode(String.valueOf(sim.getParams().getNoiseRate()*100)));
		noiseRate.setAttribute("type", "%");
		randomSeed.appendChild(xd.createTextNode(String.valueOf(sim.getParams().getRandomSeed())));
		duration.appendChild(xd.createTextNode(String.valueOf(sim.getParams().getSimDuration())));
		duration.setAttribute("type","seconds");
		transferRate.appendChild(xd.createTextNode(String.valueOf(sim.getParams().getTransferRate())));
		transferRate.setAttribute("type","bits/second");
		transmitPowerUse.appendChild(xd.createTextNode(String.valueOf(sim.getParams().getTransmitPowerUse())));
		idlePowerUse.appendChild(xd.createTextNode(String.valueOf(sim.getParams().getIdlePowerUse())));
		ttlNodes.appendChild(xd.createTextNode(String.valueOf(totalNodes)));

		setup.appendChild(name);
		setup.appendChild(noiseRate);
		setup.appendChild(randomSeed);
		setup.appendChild(duration);
		setup.appendChild(transferRate);
		setup.appendChild(transmitPowerUse);
		setup.appendChild(idlePowerUse);
		setup.appendChild(ttlNodes);

		return setup;
	}
	
	/**
	 * Takes the results and puts them into a dom.
	 * @return dom element containing the results
	 */
	private Element results(){
		Element results = xd.createElement("results");
		Element packets = xd.createElement("packets");
		Element totalPacketsTx = xd.createElement("totalSent");
		Element totalPacketsRx = xd.createElement("totalRecieved");
		Element avgHops = xd.createElement("hopsPerPacket");
		Element ttlHops = xd.createElement("totalHops");
		Element minHop = xd.createElement("minPacketHops");
		Element avgD = xd.createElement("distPerPacket");
		Element ttlD = xd.createElement("totalDist");
		Element minD = xd.createElement("minPacketDist");
		Element avgT = xd.createElement("timePerPacket");
		Element ttlT = xd.createElement("totalTime");
		Element minT = xd.createElement("minPacketTime");
		Element packetLoss = xd.createElement("loss");packetLoss.setAttribute("type", "%");
		Element frames = xd.createElement("frames");
		Element totalFrameTx = xd.createElement("totalSent");
		Element totalFrameRx = xd.createElement("totalRecieved");
		//Element fLoss = xd.createElement("loss");fLoss.setAttribute("type","%");
		Element fSize = xd.createElement("totalSent");fSize.setAttribute("type","bytes");
		Element trans = xd.createElement("transmitting");
		Element timeTx = xd.createElement("total");timeTx.setAttribute("type", "seconds");
		Element timeTxPerNode = xd.createElement("timePerNode");timeTxPerNode.setAttribute("type", "seconds");
		Element percTxTransmit = xd.createElement("percent");percTxTransmit.setAttribute("type","%");
		Element power = xd.createElement("power");
		Element powerTx = xd.createElement("total");
		Element powerTxPerNode = xd.createElement("perNode");

		totalPacketsTx.appendChild(xd.createTextNode(String.valueOf(totalPacketsSent)));
		totalPacketsRx.appendChild(xd.createTextNode(String.valueOf(totalPacketsTerminated)));
		packetLoss.appendChild(xd.createTextNode(String.valueOf(percentLoss)));
		avgHops.appendChild(xd.createTextNode(String.valueOf(averageHopLength)));
		ttlHops.appendChild(xd.createTextNode(String.valueOf(totalHops)));
		minHop.appendChild(xd.createTextNode(String.valueOf(minHopLength)));
		avgD.appendChild(xd.createTextNode(String.valueOf(avgDist)));
		ttlD.appendChild(xd.createTextNode(String.valueOf(ttlDist)));
		minD.appendChild(xd.createTextNode(String.valueOf(minDist)));
		avgT.appendChild(xd.createTextNode(String.valueOf(avgTime)));
		ttlT.appendChild(xd.createTextNode(String.valueOf(ttlTime)));
		minT.appendChild(xd.createTextNode(String.valueOf(minTime)));

		packets.appendChild(totalPacketsTx);
		packets.appendChild(totalPacketsRx);
		packets.appendChild(packetLoss);
		packets.appendChild(avgHops);
		packets.appendChild(ttlHops);
		packets.appendChild(minHop);
		packets.appendChild(avgD);
		packets.appendChild(ttlD);
		packets.appendChild(minD);
		packets.appendChild(avgT);
		packets.appendChild(ttlT);
		packets.appendChild(minT);

		totalFrameTx.appendChild(xd.createTextNode(String.valueOf(totalFramesSent)));
		totalFrameRx.appendChild(xd.createTextNode(String.valueOf(totalFramesRec)));
		//fLoss.appendChild(xd.createTextNode(String.valueOf(frameLoss)));
		fSize.appendChild(xd.createTextNode(String.valueOf(totalBytes)));

		frames.appendChild(totalFrameTx);
		frames.appendChild( makeResultElement( "totalProtocolSent", "bytes", totalProtocolSent ) );
		frames.appendChild( makeResultElement( "totalUserSent", "bytes", totalUserSent ) );
		frames.appendChild( makeResultElement( "protocolOverhead", "%", totalProtocolSent / (double) totalBytes ) );
		frames.appendChild(fSize);
		frames.appendChild(totalFrameRx);
		//frames.appendChild(fLoss);

		timeTx.appendChild(xd.createTextNode(String.valueOf(totalSendTime)));
		timeTxPerNode.appendChild(xd.createTextNode(String.valueOf(totalSendTime/totalNodes)));
		percTxTransmit.appendChild(xd.createTextNode(String.valueOf(totalSendTime/sim.getParams().getSimDuration())));

		powerTx.appendChild(xd.createTextNode(String.valueOf(powerUsage)));
		powerTxPerNode.appendChild(xd.createTextNode(String.valueOf(powerUsage/totalNodes)));

		trans.appendChild(timeTx);
		trans.appendChild(timeTxPerNode);
		trans.appendChild(percTxTransmit);

		power.appendChild(powerTx);
		power.appendChild(powerTxPerNode);

		results.appendChild(packets);
		results.appendChild(frames);
		results.appendChild(power);
		results.appendChild(trans);

		return results;
	}

	/**
	 * Makes a result dom element from name, type and value.
	 * 
	 * @param name string of tag name
	 * @param type type variable
	 * @param value inner value of the tag
	 * @return dom element
	 */
	private Element makeResultElement( String name, String type, Object value ) {
		Element ret = xd.createElement( name );
		if ( type != null )
			ret.setAttribute( "type", type );
		ret.appendChild( xd.createTextNode( value.toString() ) );
		return ret;
	}
}
