package Sim;

import Sim.Routing.FloodingRoutingAlgorithm;
import junit.framework.TestCase;

/**
 * PacketWatcherTest tests PacketWatcher functionality.
 *
 * @author Jason Winnebeck
 */
@SuppressWarnings( { "FieldCanBeLocal" } )
public class PacketWatcherTest extends TestCase {

	private Simulation    sim;
	private Packet[]      packets;
	private PacketWatcher watcher;

	private int arrived;
	private int created;
	private int outstanding;
	private int lost;

	@Override
	protected void setUp() throws Exception {
		super.setUp();

		watcher = new PacketWatcher( 3.0 );

		SimulationParameters params = new SimulationParameters( 0, 0, 100000.0, 0.0, 100.0, 200 );

		sim = new Simulation( params );

		makeNode( 0,  0, 0 );
		makeNode( 1, 10, 0 );
		makeNode( 2, 20, 0 );
		makeNode( 3, 30, 0 );
		makeNode( 4, 40, 0 );

		packets = new Packet[] {
				new Packet( 0, 4, 100, watcher ),
				new Packet( 4, 0, 200, watcher ),
				new Packet( 2, 3, 50, watcher ),
				new Packet( 3, 4, 10, watcher )
		};

		arrived = created = outstanding = lost = 0;
	}

	private Node makeNode( int addr, double x, double y ) {
		LocationManager lm = new FixedLocationManager( new Location( x, y ) );
		FloodingRoutingAlgorithm alg = new FloodingRoutingAlgorithm();
		alg.setSimulation( sim );
		Node ret = new Node( addr, sim, lm, alg, 11.0, new NodeMonitor() );
		sim.getWorld().addNode( ret );
		return ret;
	}

	/**
	 * Tests PacketWatcher through direct manipulation.
	 */
	public void testDirect() {
		checkWatcher();

		watcher.packetOriginated( packets[0], 10 );
		++created; ++outstanding;
		checkWatcher();

		watcher.updateLost( 10 );
		checkWatcher();

		watcher.updateLost( 11 );
		checkWatcher();

		watcher.packetArrived( packets[0], 12, false );
		++arrived; --outstanding;
		checkWatcher();

		watcher.packetOriginated( packets[1], 15 );
		++created; ++outstanding;
		checkWatcher();

		watcher.packetArrived( packets[1], 20, false );
		--outstanding; ++lost;
		checkWatcher();

		watcher.packetArrived( packets[1], 20.1, true );
		checkWatcher();

		watcher.updateLost( 20.2 );
		checkWatcher();

		watcher.packetOriginated( packets[2], 30 );
		++created; ++outstanding;
		checkWatcher();

		watcher.updateLost( 35 );
		--outstanding; ++lost;
		checkWatcher();

		watcher.packetArrived( packets[2], 36, false );
		checkWatcher();
	}

	/**
	 * Tests the watcher on a real network.
	 */
	public void testRealNetwork() {
		checkWatcher();
		for ( Packet packet : packets ) {
			send( packet );

			while ( sim.next() ) {
			}

			assertTrue( packet.hasArrived() );
		}

		arrived = 4;
		created = 4;
		outstanding = 0;
		lost = 0;

		checkWatcher();
	}

	public void testBrokenNetwork() {
		checkWatcher();
		makeNode( 100, 100, 100 );

		Packet packet = new Packet( 0, 100, 20, watcher );
		send( packet );

		while ( sim.next() ) {}
		assertFalse( packet.hasArrived() );

		created = 1; outstanding = 1;
		checkWatcher();

		watcher.updateLost( sim.getTime() + 3.0 );
		lost = 1; outstanding = 0;
		checkWatcher();
	}

	public void testRecurringEvent() {
		checkWatcher();
		makeNode( 100, 100, 100 );

		Packet packet = new Packet( 0, 100, 20, watcher );
		send( packet );
		watcher.scheduleRecurringLossUpdate( sim, 1.0 );

		while ( sim.getTime() < 5.0 ) {
			sim.next();
		}
		assertFalse( packet.hasArrived() );

		created = 1; lost = 1;
		checkWatcher();
	}

	private void checkWatcher() {
		assertEquals( arrived,     watcher.getArrived()     );
		assertEquals( created,     watcher.getCreated()     );
		assertEquals( outstanding, watcher.getOutstanding() );
		assertEquals( lost,        watcher.getLost()        );
	}

	private void send( Packet packet ) {
		sim.getWorld().getNode( packet.getSource() ).sendNewPacket( packet );
	}
}